"""
A series of constants useful for acquiring the player's animations.
"""

POWER_UP_OFFSETS = [0x50, 0x54, 0x54, 0x00, 0x50, 0x40, 0x44]
PLAYER_FRAME_START = 0x3AC61
PLAYER_FRAME_PAGE_OFFSET = 0x3AE47
PLAYER_SUIT_PAGE_OFFSET = 0x3AEA8
PLAYER_FRAMES = 81
PLAYER_POWER_UPS = 7
SPRITES_PER_FRAME = 6
PLAYER_POWER_UPS_PALETTE_COUNT = 8
PLAYER_POWER_UPS_PALETTES = 0x10539
ANIMATION_WIDTH = 3
ANIMATION_HEIGHT = 2
